/*  "Not the blue wire!"
 *  version 1.0, August 28th, 2013
 *
 *  Copyright (C) 2013 Jari Komppa
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty.  In no event will the authors be held liable for any damages
 *  arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, subject to the following restrictions:
 *
 *  1. The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 *  2. Altered source versions must be plainly marked as such, and must not be
 *     misrepresented as being the original software.
 *  3. This notice may not be removed or altered from any source distribution.
 *
 * Jari Komppa http://iki.fi/sol/
 *
 *************************************
 * 
 */

#include "toolkit.h"
#include "soloud.h"
#include "soloud_wav.h"

#define TITLE "Not the blue wire!"

// How big a tick difference is considered 'time warp', i.e. skip the time
// (to avoid physics blowing up)
#define TICK_TIMEWARP 1000
// Physics ticks per second
#define PHYSICS_FPS 100

//#define USE_PERFCOUNTERS

int tex_logo, tex_font;
int gKeyState = 0;
int gLastTick = 0;

SoLoud::Soloud gSoloud;
SoLoud::Wav gBasebeat, gPanic, gGameover, gPoint;
int panichandle;


void handle_key(int keysym, int down)
{
    switch(keysym)
    {
    case SDLK_ESCAPE:
        if (!down)
        {
			SoLoud::sdl_deinit(&gSoloud);
            exit(0);
        }
        break;
    case SDLK_UP:
    case SDLK_w:
        if (down) gKeyState |= KEY_UP; else gKeyState &= ~KEY_UP;
        break;
    case SDLK_DOWN:
    case SDLK_s:
        if (down) gKeyState |= KEY_DOWN; else gKeyState &= ~KEY_DOWN;
        break;
    case SDLK_LEFT:
    case SDLK_a:
        if (down) gKeyState |= KEY_LEFT; else gKeyState &= ~KEY_LEFT;
        break;
    case SDLK_RIGHT:
    case SDLK_d:
        if (down) gKeyState |= KEY_RIGHT; else gKeyState &= ~KEY_RIGHT;
        break;
    case SDLK_RCTRL:
    case SDLK_LCTRL:
        if (down) gKeyState |= KEY_FIRE;
        break;
    }
}

void process_events()
{
    SDL_Event event;

    while (SDL_PollEvent(&event)) 
    {
        switch (event.type) 
        {
        case SDL_KEYDOWN:
            handle_key(event.key.keysym.sym, 1);
			// If a key is pressed, report it to the widgets
			gUIState.keyentered = event.key.keysym.sym;
			gUIState.keymod = event.key.keysym.mod;
			// if key is ASCII, accept it as character input
			if ((event.key.keysym.unicode & 0xFF80) == 0)
				gUIState.keychar = event.key.keysym.unicode & 0x7f;				
            break;
        case SDL_KEYUP:
            handle_key(event.key.keysym.sym, 0);
            break;
        case SDL_MOUSEMOTION:
			// update mouse position
			gUIState.mousex = (event.motion.x * (float)DESIRED_WINDOW_WIDTH) / (float)gScreenWidth;
			gUIState.mousey = (event.motion.y * (float)DESIRED_WINDOW_HEIGHT) / (float)gScreenHeight;
            break;
		case SDL_MOUSEBUTTONDOWN:
			// update button down state if left-clicking
			if (event.button.button == 1)
            {
				gUIState.mousedown = 1;
			    gUIState.mousedownx = (event.motion.x * (float)DESIRED_WINDOW_WIDTH) / (float)gScreenWidth;
			    gUIState.mousedowny = (event.motion.y * (float)DESIRED_WINDOW_HEIGHT) / (float)gScreenHeight;
            }
            if (event.button.button == 4)
            {
                gUIState.scroll = +1;
            }
            if (event.button.button == 5)
            {
                gUIState.scroll = -1;
            }
			break;
		case SDL_MOUSEBUTTONUP:
			// update button down state if left-clicking
			if (event.button.button == 1)
				gUIState.mousedown = 0;
			break;
        case SDL_QUIT:
            SDL_Quit();
            exit(0);
            break;
        case SDL_VIDEORESIZE:
            gScreenWidth = event.resize.w;
            gScreenHeight = event.resize.h;
            initvideo(0);
            break;

        }
    }
}

void physics_tick(int tick)
{
}

int gStartTick = 0;

#define WIRES 37
int sockety[WIRES];
int destsocket[WIRES];
int midsocket[WIRES];
int goodsocket, goodout;
int level;
int levelwires;
int selected;
int gameover = 2;
int timeleft = 1;
int score = 0;

void init(int level);

static void draw_screen_ingame()
{
    int tick = SDL_GetTicks();

	if (gStartTick == 0)
		gStartTick = tick;

    int i, j;    

    if (tick - gLastTick > TICK_TIMEWARP) 
        gLastTick = tick;

    if (gLastTick >= tick)
    {
        SDL_Delay(1);
        return;
    }

    while (gLastTick < tick)
    {
        physics_tick(gLastTick);

        gLastTick += 1000 / PHYSICS_FPS;
    }

    ////////////////////////////////////
    // Rendering
    ////////////////////////////////////

    glClearColor(0.1,0.1,0.2,1.0);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	char temp[80];
	if (!gameover)
	{
		timeleft = (10 * 1000) - (tick - gStartTick);
		if (timeleft < 0)
			timeleft = 0;
		if (timeleft < 7500)
			gSoloud.setVolume(panichandle, (1.0f - (timeleft / 7500.0f)) * 2);
	}
	sprintf(temp,"00:00:%02d.%02d", timeleft / 1000, (timeleft / 10) % 100);
	quickfont_drawstring(tex_font,temp,8*4,0,0xff00ff00,1,8);

	sprintf(temp,"Level %d", level);
	quickfont_drawstring(tex_font,temp,8*6,100,0xff00ff00,1);
	int tl = sprintf(temp,"Score %d", score);
	quickfont_drawstring(tex_font,temp,800-8*(8+tl),100,0xff00ff00,1);
	quickfont_drawstring(tex_font,"Not the blue wire! by Jari Komppa for Ludum Dare 27 ('10 seconds') http://iki.fi/sol/",8*6,600-16,0xff00ff00,1);
	quickfont_drawstring(tex_font,"Press ESC at any time to quit.",8*6,600-32,0xff00ff00,1);

	drawrect(100,150,600,20,0xff3f3f3f);

		           // 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 
	char sockets[] = "X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X X";
	char gsockets[]= "                                                                         ";
	sockets[goodsocket * 2] = ' ';
	gsockets[goodsocket * 2] = 'O';
	sockets[levelwires * 2] = 0;
	quickfont_drawstring(tex_font,sockets,108,152,0xffff0000,1);
	quickfont_drawstring(tex_font,gsockets,108,152,0xff00ff00,1);

	if (!gameover)
	if (tick % 1000 > 200 && level == 0)
	{
		quickfont_drawstring(tex_font, " CUT HERE ->", 0, 452, 0xffffffff,1);
		quickfont_drawstring(tex_font, "<- CUT HERE", 704, 452, 0xffffffff,1);
	}

	drawrect(100,450,600,20,0xff5f5f5f);

	if (gUIState.mousedown && !gameover)
	{
		selected = -1;
	}

	for (i = 0; i < levelwires; i++)
	{
		if (rect_rect_collide(gUIState.mousex, gUIState.mousey, 
							  gUIState.mousex, gUIState.mousey,
							  106+i*16,450, 
							  106+i*16+10,450+20) && !gameover)
		{
			if (gUIState.mousedown)
			{
				selected = i;
				drawrect(106+i*16,450,10,20,0xffffffff);
			}
			else
			drawrect(106+i*16,450,10,20,0xffff7f7f);
		}
		else
		{
			drawrect(106+i*16,450,10,20,0xff7f7f7f);
		}

	}

	if (gameover == 1)
	{
		char esockets[]= "                                                                         ";
		esockets[goodout * 2] = 'O';
		quickfont_drawstring(tex_font,esockets,106,452,0xff00ff00,1);
	}

	float tlf = timeleft / 10000.0f;
	tlf = tlf * tlf * tlf * tlf * tlf * tlf;

	glEnable(GL_BLEND); glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glLineWidth(3);
	glColor3f(0.5,0.5,1);

	for (i = 0; i < levelwires; i++)
	{
		float midx = (106+6+midsocket[i]*16 + tlf*sin((tick+i*100)*0.00234)*20);
		float midy = (170 + (sockety[i] + 6) * (280/(WIRES+8)));

		float dx = midx - gUIState.mousex;
		float dy = midy - gUIState.mousey;
		float d = sqrt(dx*dx+dy*dy);
		if (d < 200 && !gameover)
		{
			midx += dx*dx*dx * (200 - d) / 7000000;
			midy += dy*dy*dy * (200 - d) / 7000000;
		}


		glLineWidth(5);
		glColor3f(0,0,0);
		glBegin(GL_LINES);
		glVertex2f(106+6+i*16,170);
		glVertex2f(106+6+i*16,180);
		glVertex2f(106+6+i*16,180);
		glVertex2f(midx,midy);
		glVertex2f(midx,midy);
		glVertex2f(106+6+destsocket[i]*16,440);
		glVertex2f(106+6+destsocket[i]*16,440);
		glVertex2f(106+6+destsocket[i]*16,450);
		glEnd();

		glLineWidth(3);
		glColor3f(0.5,0.5,1);
		glBegin(GL_LINES);
		glVertex2f(106+6+i*16,170);
		glVertex2f(106+6+i*16,180);
		glVertex2f(106+6+i*16,180);
		glVertex2f(midx,midy);
		glVertex2f(midx,midy);
		glVertex2f(106+6+destsocket[i]*16,440);
		glVertex2f(106+6+destsocket[i]*16,440);
		glVertex2f(106+6+destsocket[i]*16,450);
		glEnd();
	}

	if (timeleft == 0)
	{
		// game over
		gameover = 1;
		gSoloud.play(gGameover,2);
		gSoloud.fadeVolume(panichandle, gSoloud.getVolume(panichandle), 0, 0.5);
		selected = -1;
		gUIState.mousedown = 0;
		timeleft = 1; // kludge
	}

	if (!gUIState.mousedown && selected != -1 && !gameover)
	{
		if (selected == goodout)
		{
			level++;
			init(level);
			gSoloud.play(gPoint);
			gSoloud.fadeVolume(panichandle, gSoloud.getVolume(panichandle), 0, 0.5);
			gStartTick = tick;
			score += timeleft;
		}
		else
		{
			// game over
			gameover = 1;
			gSoloud.play(gGameover,2);
			gSoloud.fadeVolume(panichandle, gSoloud.getVolume(panichandle), 0, 0.5);
			selected = -1;
			gUIState.mousedown = 0;
		}
	}

	if (gameover)
	{
		drawrect(0,0,800,600,0x7f7f7f7f);
		if (gameover == 1)
		{
			if (tick % 500 > 100)
			{
				quickfont_drawstring(tex_font,"!!! BOOM !!!",8*2+8,250+8,0xff000000,1,8);
				quickfont_drawstring(tex_font,"!!! BOOM !!!",8*2,250,0xffff0000,1,8);
			}
		}
		else
		{
			quickfont_drawstring(tex_font,"Not the blue wire!",8*6+8,250+8,0xff000000,1,5);
			quickfont_drawstring(tex_font,"Not the blue wire!",8*6,250,0xff7f7fff,1,5);
		}

		drawrect(200,500,400,50,0xff000000);

		if (gUIState.mousedown)
			selected = -1;

		if (rect_rect_collide(gUIState.mousex, gUIState.mousey, 
							  gUIState.mousex, gUIState.mousey,
							  200,500, 
							  200+400,500+50))
		{
			if (gameover == 1)
			quickfont_drawstring(tex_font,"Restart",320,500,0xffffffff,1,3);
			else
			quickfont_drawstring(tex_font," Start ",320,500,0xffffffff,1,3);
			if (gUIState.mousedown)
				selected = 1;
		}
		else
		{
			if (gameover == 1)
			quickfont_drawstring(tex_font,"Restart",320,500,0xffffffff,1,3);
			else
			quickfont_drawstring(tex_font," Start",320,500,0xff7f7f7f,1,3);
		}

		if (!gUIState.mousedown && selected != -1)
		{
			gameover = 0;
			level = 0;
			init(level);
			gStartTick = tick;
			score = 0;
		}
	}


#ifdef USE_PERFCOUNTERS

	render_perfcounters(tex_font);

#endif
    SDL_Delay(10);
	SDL_GL_SwapBuffers();
}



void draw_screen()
{
    draw_screen_ingame();
}

void swapsockets1(int a, int b)
{
	int t = destsocket[a];
	destsocket[a] = destsocket[b];
	destsocket[b] = t;
}

void swapsockets2(int a, int b)
{
	int t = midsocket[a];
	midsocket[a] = midsocket[b];
	midsocket[b] = t;
}

void swapsockets3(int a, int b)
{
	int t = sockety[a];
	sockety[a] = sockety[b];
	sockety[b] = t;
}

void init(int level)
{
	levelwires = level*2 + 5;
	if (levelwires > WIRES)
		levelwires = WIRES;
	selected = -1;
	int i;
	for (i = 0; i < WIRES; i++)
	{
		sockety[i] = i;
		destsocket[i] = i;
		midsocket[i] = i;
	}

	goodsocket = gVisualRand.genrand_int31() % levelwires;
	goodout = goodsocket;

	for (i = 0; i < (level / 2) + 1; i++)
	{
		swapsockets1(gVisualRand.genrand_int31() % levelwires, gVisualRand.genrand_int31() % levelwires);
		swapsockets2(gVisualRand.genrand_int31() % levelwires, gVisualRand.genrand_int31() % levelwires);
		swapsockets3(gVisualRand.genrand_int31() % levelwires, gVisualRand.genrand_int31() % levelwires);
	}

	swapsockets1(goodsocket, gVisualRand.genrand_int31() % levelwires);

	goodout = destsocket[goodsocket];

//	for (i = 0; i < levelwires; i++)
//		midsocket[i] = (destsocket[i] + i) / 2;

	glEnable(GL_LINE_SMOOTH);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
}

int main (int argc, char** args)
{
	level = 0;

    gVisualRand.init_genrand(0xc0cac01a);

    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE) < 0) 
    {
        fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
        SDL_Quit();
        exit(0);
    }

	SoLoud::sdl_init(&gSoloud);

	gBasebeat.load("basebeat.ogg");
	gBasebeat.setLooping(1);
	gPanic.load("panic.ogg");
	gPanic.setLooping(1);
	gGameover.load("gameover.ogg");
	gPoint.load("point.ogg");

	int basehandle = gSoloud.play(gBasebeat);
	panichandle = gSoloud.play(gPanic,0);
	gSoloud.setProtectChannel(basehandle, 1);
	gSoloud.setProtectChannel(panichandle, 1);

    initvideo(argc);

	init(100);

    tex_font = load_texture("fona.png");
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST); // Linear Filtering


	// For imgui - Enable keyboard repeat to make sliders more tolerable
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
	// For imgui - Enable keyboard UNICODE processing for the text field.
	SDL_EnableUNICODE(1);
   
    // set window title
    SDL_WM_SetCaption(TITLE " - http://iki.fi/sol/", NULL);

    // hide cursor

//	SDL_SetCursor(load_cursor("cursor_scissors.png",0,0));
    
    while (1) 
    {
        process_events();
        draw_screen();
    }

    return 0;
}
